import json
import time
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, Sequence

from pybaked import protocol

bake_parser = ArgumentParser()
bake_parser.add_argument("package", help="Package to bake")
bake_parser.add_argument(
    "-o", "--output", help="Output file name", default=None, required=False
)
bake_parser.add_argument(
    "-H",
    "--hash",
    help="Hash content and write it to file",
    action="store_true",
    default=False,
)
bake_parser.add_argument(
    "--no-colors", help="Don't color output", action="store_true", default=False
)
bake_parser.add_argument(
    "-m",
    "--metadata",
    help="JSON encoded metadata",
    required=False,
    default="{}",
)
bake_parser.add_argument(
    "-M",
    "--metadata-file",
    help="JSON encoded metadata file path",
    required=False,
    default=None,
)

read_parser = ArgumentParser()
read_parser.add_argument(
    "baked_package",
    help="Baked package file",
)
read_parser.add_argument(
    "-m",
    "--module",
    help="Read specific module source",
    required=False,
    default=None,
)
read_parser.add_argument(
    "--no-colors", help="Don't color output", action="store_true", default=False
)

USE_COLORS = True


def color(s: str, code: int) -> str:
    if not USE_COLORS:
        return str(s)

    s = str(s)

    if "\x1b[0m" in s:
        s = s.replace("\033[0m", f"\033[{code}m")

    return f"\033[{code}m{s}\033[0m"


def red(s: Any) -> str:
    return color(s, 31)


def green(s: Any) -> str:
    return color(s, 32)


def yellow(s: Any) -> str:
    return color(s, 33)


def cyan(s: Any) -> str:
    return color(s, 36)


def blue(s: Any) -> str:
    return color(s, 34)


def purple(s: Any) -> str:
    return color(s, 35)


def bake():
    args = bake_parser.parse_args()

    if args.no_colors:
        global USE_COLORS
        USE_COLORS = False

    package_path = Path(args.package)

    if not package_path.is_dir():
        print(red(f"Package {yellow(args.package)} not found"))
        return -1

    metadata = json.loads(args.metadata)

    if args.metadata_file is not None:
        mf = Path(args.metadata_file)
        if not mf.is_file():
            print(
                red(
                    f"Metadata file {yellow(args.metadata_file)} is not "
                    f"exists or is not a file"
                )
            )
            return -2

        metadata = json.loads(mf.read_text())

    from pybaked import BakedMaker

    baker = BakedMaker.from_package(package_path, args.hash, metadata)

    print(
        cyan(
            f"Baking {yellow(args.package)} as {yellow(args.output or package_path.name)}..."
        ),
        flush=True,
        end="\r",
    )

    filename = baker.file(args.output or package_path)

    print(
        green(
            f"Baked package {yellow(args.package)} into file {cyan(filename)}"
        )
    )

    return 0


def format_metadata(metadata: dict[str, Any]) -> str:
    lines = []
    for key, value in metadata.items():
        if isinstance(value, dict):
            value = "\n\t" + format_metadata(value).replace("\n", "\n\t")

        elif isinstance(value, list):
            value = "\n" + "\n".join((f"\t- {e}" for e in value))

        elif isinstance(value, str):
            value = green(repr(value))

        elif isinstance(value, (int, float)):
            value = blue(value)

        elif isinstance(value, bool):
            value = green(value) if value else red(value)

        else:
            value = yellow(repr(value))

        key = yellow(key) + ": "

        lines.append(f"{key}{value}")

    return "\n".join(lines)


def format_modules(modules: Sequence[str]) -> str:
    try:
        from asciitree import LeftAligned
        from asciitree.drawing import BoxStyle, BOX_LIGHT

        render = LeftAligned(draw=BoxStyle(gfx=BOX_LIGHT))

        tree = {}
        for module in modules:
            package = tree
            parts = module.split(".")

            for deep, part in enumerate(parts):
                if deep == len(parts) - 1:
                    real = module.split(".", 1)[1]
                    part += f" ({yellow(real)})"
                package = package.setdefault(purple(part), {})

        return green(render(tree))

    except ImportError:
        pass

    modules_formatted: dict[str, str] = {}

    for name in modules:
        modules_formatted["/".join(map(purple, name.split(".")[1:]))] = yellow(
            ".".join(name.split(".")[1:])
        )

    return "\n".join(
        f"\t- {pretty} ({real})" for pretty, real in modules_formatted.items()
    )


def read():
    args = read_parser.parse_args()

    if args.no_colors or args.module is not None:
        global USE_COLORS
        USE_COLORS = False

    if args.baked_package.endswith(".py"):
        print(
            red(
                f"Expected baked package. But given python module {yellow(args.baked_package)}"
            )
        )
        return -2

    baked_package = args.baked_package

    if not args.baked_package.endswith(protocol.EXTENSION):
        package_name = args.baked_package
        baked_package = args.baked_package + protocol.EXTENSION
    else:
        package_name = args.baked_package[: -len(protocol.EXTENSION)]

    package_path = Path(baked_package)

    if not package_path.is_file():
        print(red(f"Package {yellow(baked_package)} not found"))
        return -1

    from pybaked import BakedReader

    if args.module is None:
        print(cyan(f"Reading {yellow(baked_package)}..."), flush=True, end="\r")

    reader = BakedReader(package_path)
    modules = reader.modules_dict
    packages = reader.packages[1:]

    if args.module is not None:
        module = package_name + "." + args.module

        if module not in modules:
            print(f"Module {args.module} not found in {baked_package}")
            return -3

        source = reader.read_specific(modules[module])
        print(source.decode())

        return 0

    print(
        green(f"Package {yellow(baked_package)} read successfully"), end="\n\n"
    )

    print(green(f"Hash supported: {yellow(reader.hash_match is not None)}"))
    if reader.hash_match is not None:
        color = green if reader.hash_match else red
        print(green(f"Hash matched: {color(reader.hash_match)}"))

    print(green(f"Creation date: {blue(str(reader.created))}"))
    print(green("Metadata:"))
    print(format_metadata(reader.metadata), "\n")
    print(green(f"Subpackages ({blue(len(packages))}):"))
    print(
        "\n".join(
            "\t- " + yellow(".").join(map(purple, package.split(".")[1:]))
            for package in packages
        ),
        end="\n\n",
    )
    print(green(f"Modules ({blue(len(modules))}):"))
    print(format_modules(list(modules.keys())))



import logging
from datetime import datetime
from functools import lru_cache
from pathlib import Path
from typing import Any

from . import protocol, BakedMaker

logger = logging.getLogger(__name__)


class BakedReader:
    def __init__(self, path: str | Path):
        if not isinstance(path, Path):
            path = Path(path)

        self._path = path.absolute()

        if not self._path.is_file() or not self._path.name.endswith(
            protocol.EXTENSION
        ):
            raise ValueError(f"Baked file does not exist: {self._path}")

        self._file = self._path.open("rb")

        data = self._read_next()

        if not data:
            raise ValueError(
                "Cannot decode baked file: creation timestamp not found"
            )

        self._created = protocol.deserialize(data)
        logger.debug(f"Read creation date from {path} => {self._created}")

        data = self._read_next()

        if not data:
            raise ValueError("Cannot decode baked file: metadata not found")

        self._metadata = protocol.deserialize(data)
        logger.debug(f"Read metadata from {path} => {self._metadata}")

        self._modules_offset = self._file.tell()

    def _read_next(self) -> bytes | None:
        """
        Read next data from the file
        """
        length_bytes = self._file.read(8)
        if len(length_bytes) != 8:
            return None

        length = int.from_bytes(length_bytes, "little")

        return self._file.read(length)

    def read_specific(self, offset: int) -> bytes:
        self._file.seek(offset)

        logger.debug(f"Reading data at {offset} from {self._path}")

        return self._read_next()

    @property
    def path(self):
        return self._path

    @property
    @lru_cache
    def hash_match(self) -> bool | None:
        """
        Match hash written in metadata with real hash

        :return: None if no hash is present in metadata, bool - hash match
        """
        if "--fh" not in self._metadata:
            return

        return self.real_hash == self._metadata["--fh"]

    @property
    @lru_cache
    def real_hash(self) -> bytes:
        """
        Real content hash of the file

        :return: hash bytes
        """
        self._file.seek(self._modules_offset)

        return protocol.hash_fragments(self._file)

    @property
    def metadata(self) -> dict[str, Any]:
        return self._metadata.copy()

    @property
    def created(self) -> datetime:
        return self._created

    @property
    def name(self):
        return self._path.name[: -len(protocol.EXTENSION)]

    @property
    @lru_cache
    def modules_dict(self) -> dict[str, int]:
        return dict(self.modules)

    @property
    @lru_cache
    def modules(self) -> list[tuple[str, int]]:
        found_modules = []

        logger.debug(f"Reading modules from {self._path}")

        self._file.seek(self._modules_offset)
        fragments = protocol.read_fragments(self._file)

        for name, offset in fragments:
            found_modules.append((self.name + "." + name.decode(), offset))

        logger.debug(
            f"Read {len(found_modules)} modules from {self._path}",
            extra={"modules": found_modules},
        )

        return found_modules

    @property
    @lru_cache
    def packages(self):
        found_packages = []

        logger.debug(f"Defining packages for {self._path}")

        for name in self.modules_dict:
            name = name.rsplit(".", 1)[0]

            if name not in found_packages:
                found_packages.append(name)

        logger.debug(
            f"Found {len(found_packages)} packages in {self._path}",
            extra={"packages": found_packages},
        )

        return list(found_packages)

    def to_maker(self) -> BakedMaker:
        maker = BakedMaker("--fh" in self.metadata, self.metadata)

        for module_name, source_offset in self.modules_dict.items():
            maker.include_module(
                module_name[len(self.name) + 1 :].encode(),
                self.read_specific(source_offset),
            )

        return maker

    def __del__(self):
        self._file.close()



from datetime import datetime
from pathlib import Path
from typing import Any
import logging
import os
import io

from . import protocol

logger = logging.getLogger(__name__)


def write_content(buffer: io.BytesIO, content: bytes):
    logger.debug(
        f"Writing {len(content)} bytes into a buffer",
    )
    buffer.write(len(content).to_bytes(8, byteorder="little"))
    buffer.write(content)


def find_modules(package_path: Path):
    modules: list[tuple[str, str]] = []
    for dir_path, _, files in os.walk(package_path):
        dir_path = Path(dir_path)
        for module_file in filter(lambda name: name.endswith(".py"), files):
            module_path = dir_path / module_file

            module_name = ".".join(
                (module_path.relative_to(package_path)).with_suffix("").parts
            )

            modules.append((module_name, str(module_path)))

    return modules


class BakedMaker:
    logger = logger.getChild("BakedMaker")

    @classmethod
    def from_package(
        cls,
        package_path: str | Path,
        hash_content: bool = False,
        metadata: dict[str, Any] = None,
    ) -> "BakedMaker":
        """
        Lookup path for python modules and create BakedMaker instance
        including all this modules
        :param package_path: Path to package
        :param hash_content: Whether to hash the content
        :param metadata: Metadata dictionary
        :return: The created BakedMaker instance
        """
        if isinstance(package_path, str):
            package_path = Path(package_path)

        package_path = package_path.absolute()

        if not package_path.is_dir():
            raise ValueError("Package path is not a directory")

        if not (modules := find_modules(package_path)):
            raise ValueError("No modules found in package")

        instance = cls(hash_content, metadata)

        for import_name, module_file in modules:
            cls.logger.debug(
                f"Including '{module_file}' as '{import_name}' from '{package_path}'"
            )
            instance.include_module(
                import_name.encode(), Path(module_file).read_bytes()
            )

        return instance

    def __init__(
        self, hash_content: bool = False, metadata: dict[str, Any] = None
    ):
        if metadata is None:
            metadata = {}

        if not isinstance(metadata, dict):
            raise ValueError("Metadata must be a dict")

        self._hash_content = hash_content
        self._metadata = metadata

        self._fragments = protocol.Fragments()

    def get_metadata(self) -> dict[str, Any]:
        """
        Get the metadata dictionary

        :return: Metadata dictionary
        """
        return self._metadata

    def set_metadata(self, metadata: dict[str, Any] = None) -> "BakedMaker":
        """
        Set the new metadata for this maker.
        If metadata is None - metadata will be cleared

        :param metadata: New metadata to set
        :return: The same instance of BakedMaker
        """
        if metadata is None:
            metadata = {}

        self._metadata = metadata
        return self

    def update_metadata(
        self, data: dict[str, Any] = None, **values: Any
    ) -> "BakedMaker":
        """
        Update metadata for this maker with the new values.

        :param data: The dictionary that will be used to update the metadata
        :param values: Keywords that will update the metadata
        :return: The same instance of BakedMaker
        """
        if data is None:
            data = values
        else:
            data.update(values)

        self._metadata.update(data)

        return self

    def include_module(
        self, import_name: bytes, source_code: bytes
    ) -> "BakedMaker":
        """
        Include module in BakedMaker

        :param import_name: Module name if import format (example: baked_package.module_name)
        :param source_code: Module source code
        :return: The same instance of BakedMaker
        """
        self._fragments.add(
            (
                import_name,
                source_code,
            )
        )

        return self

    def _build_content(self) -> io.BytesIO:
        self.logger.debug("Started building content")
        buffer = io.BytesIO()
        creation_date = datetime.utcnow()
        write_content(
            buffer,
            protocol.serialize(
                creation_date,
            ),
        )
        self.logger.debug(
            "Creation date was written to a buffer",
            extra={"creation_date": creation_date},
        )
        if self._hash_content:
            fragments_hash = self._fragments.hash()
            self._metadata.update({"--fh": fragments_hash})
            self.logger.debug(
                "Fragments was hashed",
                extra={"hash": int.from_bytes(fragments_hash, "big")},
            )

        write_content(buffer, protocol.serialize(self._metadata))
        self.logger.debug(
            "Metadata was written to a buffer",
        )

        self._fragments.write(buffer)
        self.logger.debug("Fragments was written to a buffer")

        self.logger.debug("Building content finished")

        buffer.seek(0)
        return buffer

    def bytes(self):
        """
        Make baked package in bytes from stored content

        :return: package bytes
        """
        return self._build_content().read()

    def file(self, filename: str | Path) -> Path:
        """
        Make baked package file from stored content

        :param filename: output file name
        :return: path to file created
        """
        if isinstance(filename, str):
            filename = Path(filename)

        if not filename.name.endswith(protocol.EXTENSION):
            filename = filename.with_name(
                filename.name.split(".", 1)[0] + protocol.EXTENSION
            )

        with filename.open("wb") as f:
            f.write(self.bytes())

        return filename



import hashlib
from datetime import datetime
from typing import Any, Callable, TypeVar

EXTENSION = ".py.baked"


def pack_type(data: bytes, type_: str) -> bytes:
    return type_.encode() + b"/" + data


def unpack_type(data: bytes) -> tuple[bytes, str]:
    type_, data = data.split(b"/", 1)

    return (data, type_.decode())


def int_serialize(i: int) -> bytes:
    return pack_type(i.to_bytes(8, "little"), "int")


def int_deserialize(b: bytes) -> int:
    return int.from_bytes(b, "little")


def bool_serialize(b: bool) -> bytes:
    return pack_type(b.to_bytes(1, "little"), "bool")


def bool_deserialize(b: bytes) -> bool:
    return bool(int_deserialize(b))


def str_serialize(s: str) -> bytes:
    return pack_type(s.encode(), "str")


def str_deserialize(b: bytes) -> str:
    return b.decode()


def float_serialize(f: float) -> bytes:
    f_str = str(f)
    f_parts = f_str.split(".")

    exp = len(f_parts[1])
    exp_bytes = exp.to_bytes(1, "little")

    real = int("".join(f_parts))
    real_bytes = real.to_bytes(8, "little")

    return pack_type(exp_bytes + b"-" + real_bytes, "float")


def float_deserialize(b: bytes) -> float:
    exp_bytes, real_bytes = b.split(b"-", 1)

    exp = 10 ** int.from_bytes(exp_bytes, "little")

    real = int.from_bytes(real_bytes, "little")

    return real / exp


def list_serialize(lst: list) -> bytes:
    buffer = b""

    for i, element in enumerate(lst):
        try:
            buffer += pack_message(serialize(element))
        except TypeError as e:
            raise TypeError(f"Unsupported type for {i}") from e

    return pack_type(buffer, "list")


def list_deserialize(b: bytes) -> list:
    elements = []

    cursor = 0

    while len(length_bytes := b[cursor : cursor + 8]) > 0:
        cursor += len(length_bytes)
        length = int.from_bytes(length_bytes, "little")

        elements.append(deserialize(b[cursor : cursor + length]))

        cursor += length

    return elements


def dict_serialize(d: dict[str, Any]) -> bytes:
    buffer = b""

    for key, value in d.items():
        buffer += pack_message(key.encode())

        try:
            buffer += pack_message(serialize(value))
        except TypeError as e:
            raise TypeError(f"Unsupported value type for {key}") from e

    return pack_type(buffer, "dict")


def dict_deserialize(b: bytes) -> dict[str, Any]:
    dct: dict[str, Any] = {}

    cursor = 0

    while len(length_bytes := b[cursor : cursor + 8]) > 0:
        cursor += len(length_bytes)
        length = int.from_bytes(length_bytes, "little")

        element_key = b[cursor : cursor + length]

        cursor += len(element_key)

        length_bytes = b[cursor : cursor + 8]
        cursor += len(length_bytes)
        length = int.from_bytes(length_bytes, "little")

        dct[element_key.decode()] = deserialize(b[cursor : cursor + length])

        cursor += length

    return dct


def datetime_serialize(d: datetime) -> bytes:
    return pack_type(unpack_type(float_serialize(d.timestamp()))[0], "datetime")


def datetime_deserialize(b: bytes) -> datetime:
    return datetime.fromtimestamp(float_deserialize(b))


_types = {
    "int": (
        int_serialize,
        int_deserialize,
    ),
    "str": (str_serialize, str_deserialize),
    "float": (float_serialize, float_deserialize),
    "list": (list_serialize, list_deserialize),
    "bool": (bool_serialize, bool_deserialize),
    "dict": (dict_serialize, dict_deserialize),
    "datetime": (datetime_serialize, datetime_deserialize),
}


def deserialize(b: bytes) -> Any:
    data, type_ = unpack_type(b)

    if type_ == "bytes":
        return data

    if type_ not in _types:
        raise TypeError(f"Unsupported data type: {type_}")

    _, deserialize = _types[type_]

    return deserialize(data)


def serialize(data: Any) -> bytes:
    type_ = type(data).__name__

    if type_ == "bytes":
        return pack_type(data, "bytes")

    if type_ not in _types:
        raise TypeError(f"Unsupported data type: {type_}")

    serialize, _ = _types[type_]

    return serialize(data)


T = TypeVar("T")


def register_type(
    type_: type[T],
    serialize: Callable[[T], bytes],
    deserialize: Callable[[bytes], T],
):
    if type_ in _types:
        raise ValueError("Type already registered")
    _types[type_.__name__] = (serialize, deserialize)


def pack_message(message: bytes) -> bytes:
    return len(message).to_bytes(8, "little") + message


class Fragments:
    def __init__(self):
        self._fragments: list[tuple[bytes, bytes]] = []

    def hash(self) -> bytes:
        hash_ = hashlib.sha256()

        for name, content in self._fragments:
            hash_.update(name + content)

        return hash_.digest()

    def add(self, fragment: tuple[bytes, bytes]):
        self._fragments.append(fragment)

    def measure_offset(self, offset: int):
        return sum(
            map(lambda x: len(x[0]) + 24, self._fragments),
        ) + sum(map(lambda x: len(x[1]) + 8, self._fragments[:offset]))

    def write(self, buffer):
        for i, element in enumerate(self._fragments):
            buffer.write(pack_message(element[0]))
            buffer.write(
                pack_message(self.measure_offset(i).to_bytes(8, "little"))
            )

        for element in self._fragments:
            buffer.write(pack_message(element[1]))

    def __iter__(self):
        return iter(self._fragments)


def read_buffer(buffer) -> bytes:
    """
    Reads message from the buffer and returns it. Raises ValueError otherwise.

    MESSAGE := 8 bytes of length + bytes of message
    :param buffer: file-like object with rb mode
    :return: bytes read
    """
    length_bytes = buffer.read(8)
    if len(length_bytes) != 8:
        raise ValueError(
            "Buffer ended unexpectedly while reading length of the message"
        )

    length = int.from_bytes(length_bytes, byteorder="little", signed=False)

    data = buffer.read(length)

    if len(data) != length:
        raise ValueError(
            "Buffer ended unexpectedly while reading the message content"
        )

    return data


def read_fragments(buffer) -> list[tuple[bytes, int]]:
    """
    Reads all fragments from buffer. Returns a list of tuples (name, offset).
    Where the offset is the body position of the fragment and name is the fragment name

    :param buffer: file-like object with rb mode
    :return: tuples of fragment name and offset to its content
    """

    position = buffer.tell()

    fragments = []

    name = read_buffer(buffer)

    offset = int.from_bytes(read_buffer(buffer), "little", signed=False)
    latest_name_offset = position + offset
    next_offset = buffer.tell()

    while next_offset <= latest_name_offset:
        fragments.append((name, position + offset))

        if next_offset == latest_name_offset:
            break

        buffer.seek(next_offset)

        name = read_buffer(buffer)

        offset = int.from_bytes(read_buffer(buffer), "little", signed=False)

        next_offset = buffer.tell()

    return fragments


def hash_fragments(buffer) -> bytes:
    """
    Reads all fragments and its content from file and makes hash of it.

    :param buffer: file-like object with rb mode
    """
    hash_ = hashlib.sha256()
    for name, offset in read_fragments(buffer):
        hash_.update(name)

        buffer.seek(offset)
        content = read_buffer(buffer)
        hash_.update(content)

    return hash_.digest()



import logging
import importlib.util
import sys
import types
from functools import lru_cache
from importlib.abc import MetaPathFinder, Loader
from pathlib import Path

from pybaked import BakedReader, protocol

module_logger = logging.getLogger(__name__)


def execute_module(source: str | bytes, module: types.ModuleType):
    code = compile(source, module.__file__, "exec")
    exec(code, module.__dict__)


class BakedPathFinder(MetaPathFinder):
    logger = module_logger.getChild("BakedPathFinder")

    @lru_cache(10)
    def reader_for(self, path: Path) -> BakedReader:
        return BakedReader(path)

    def find_spec(self, fullname, path, target=...):
        # Define path entries in which finder will search baked packages
        entries = sys.path
        # Extend entries with package path
        if path is not None:
            entries.extend(path)

        for path_entry in entries:
            self.logger.debug(
                f"Lookup {path_entry} for baked package in {fullname}"
            )

            # Using more convenient interface for working with path
            baked_package = Path(path_entry)

            # Find baked package inside path entry following
            # the module name parts
            parts = fullname.split(".")
            for i, part in enumerate(parts):
                baked_package /= part

                # If this is baked package - use it and break loop
                with_suffix = baked_package.with_suffix(protocol.EXTENSION)
                if with_suffix.exists():
                    baked_package = with_suffix

                    # Define inner module name relative to
                    inner_module_name = ".".join(parts[i:])

                    self.logger.debug(
                        f"Found baked package at {baked_package} for {fullname}(inner name: {inner_module_name})"
                    )
                    break
            else:
                # If package was not found - skip this path entry
                continue

            # Defining baked package reader to search module inside
            reader = self.reader_for(baked_package)

            # If content hash is not matched to metadata hash - skip this baked package
            if reader.hash_match is False:
                self.logger.debug(
                    f"Corrupted baked package at {baked_package} - skipping"
                )
                continue

            # Define module location
            location = str(
                baked_package.joinpath(*inner_module_name.split(".")[1:])
            )

            self.logger.debug(
                f"Lookup {baked_package} for module {inner_module_name}"
            )

            # If it is a normal module - add .py suffix (just in case)
            if inner_module_name in reader.modules_dict:
                location += ".py"

            # If module not found, and it is not a package
            # inside with this name - skip this baked package
            elif inner_module_name not in reader.packages:

                self.logger.debug(
                    f"Module {inner_module_name} not found in {baked_package} -"
                    f" abort searching for another packages"
                )
                return None

            self.logger.debug(
                f"Module {inner_module_name} found in {baked_package} - proceed loading"
            )

            # Build spec for module
            return importlib.util.spec_from_file_location(
                fullname,
                location,
                loader=BakedLoader(reader, inner_module_name),
            )
        return None


class BakedLoader(Loader):
    logger = module_logger.getChild("BakedLoader")

    def __init__(self, reader: BakedReader, inner_module_name: str):
        self.reader = reader
        self.inner_module_name = inner_module_name

    def create_module(self, spec):
        return types.ModuleType(spec.name)

    def exec_module(self, module):
        self.logger.debug(
            f"Loading module {module.__name__} from {self.reader.path}"
        )
        # Define what module will be read from reader.
        # If module is the package then it may be
        # package_name + .__init__ if exists module with that name
        module_name = self.inner_module_name

        # If module is the package - package must be equal to module name
        if module_name in self.reader.packages:
            module.__package__ = module.__name__

            # If package has __init__ module - load it
            if module_name + ".__init__" in self.reader.modules_dict:
                module_name += ".__init__"
        else:
            module.__package__ = module.__name__.rsplit(".", 1)[0]

        # Location module inside the baked package
        # (first element is the name of the package)
        module_location_inside = module_name.split(".")[1:]

        # Define path of the module file
        module.__file__ = (
            str(self.reader.path / Path(*module_location_inside)) + ".py"
        )

        # Define package module resolution path
        module.__path__ = [self.reader.path]

        # If module not found in the package - leaving from loader
        if module_name not in self.reader.modules_dict:
            self.logger.debug(
                f"Module {module_name}({module.__name__}) not found in {self.reader.path}"
            )
            return

        # Offset of the module source in the baked package file
        source_offset = self.reader.modules_dict[module_name]

        # Source of the module
        source = self.reader.read_specific(source_offset)

        self.logger.debug(f"Executing module {module_name}({module.__name__})")

        # Compile and execute module code
        execute_module(source, module)

        self.logger.debug(
            f"Module {module_name}({module.__name__}) successfully executed"
        )

        # Install metadata read by reader to all root modules of the baked package
        # Use it to recognize baked modules in code
        if module_name.split(".", 1)[0] == module_name.rsplit(".", 1)[0]:
            self.logger.debug(
                f"Installing metadata for module {module_name}({module.__name__})"
            )
            module.__baked_metadata__ = self.reader.metadata


def init():
    # If already initiated - do nothing
    for finder in sys.meta_path:
        if isinstance(finder, BakedPathFinder):
            return

    sys.meta_path.append(BakedPathFinder())
